import hashlib
import hmac
import jsonpickle
import os
import random
import requests
import string
import time

try:
    import configparser as parser
except ImportError:
    import ConfigParser as parser

from pyvdp.exceptions import (VisaAccessDeniedError, VisaDuplicateTransactionError, VisaGeneralError,
                              VisaMessageValidationError, VisaNotFoundError, VisaTimeoutError, VisaUnauthenticatedError)


class VisaRequest(object):
    """Constructs a request to Visa API to provided resource, method and http verb.

    After request is constructed, it shall be submitted with `send()` method, which returns a `requests.Response`
    object.

    This class is not assumed to be instantiated on its own (consider it abstract). Instead, every VISA API
    implementation must provide its own request class, inheriting from VisaRequest and providing a
    corresponding VISA API resource name as a `resource` argument value (see visa.visadirect.request as an example)

    :param str resource: **Required**. VISA API resource name
    :param str api: **Required**. VISA API api name
    :param str method: **Required**. VISA API endpoint method
    :param str http_verb: **Required**. VISA API HTTP verb (GET, POST, PUT)
    :param str query_string: **Conditional**. Query string to append to API endpoint
    :param object data: **Conditional**. Data payload for POST requests, VisaTransacton object
    :param str auth_method: **Conditional**. Authentication method. Possible values are:
            **ssl** - for certificate-based authentication (default) or **token** - for x-pay-token authentication
    :param dict headers: **Optional**. Additional headers.
    """
    # This is where stuff lives
    BASE_DIR = os.path.dirname(__file__)

    # Configuration management
    config = parser.ConfigParser()
    config_file = os.path.join(BASE_DIR, 'configuration.ini')

    if os.path.exists(config_file):
        config.read(config_file)
    else:
        # use example config as a stub for tests
        config.read(os.path.join(BASE_DIR, 'configuration.ini.example'))

    # VDP HTTP codes mapped to exceptions
    # (kudos to http://codereview.stackexchange.com/questions/155350/pyvdp-python-library-for-visa-developer-program)
    ERROR_CODES = {
        202: VisaTimeoutError,
        303: VisaDuplicateTransactionError,
        400: VisaMessageValidationError,
        401: VisaUnauthenticatedError,
        403: VisaAccessDeniedError,
        404: VisaNotFoundError
    }

    def __init__(self, resource, api, method, http_verb, query_string='', data='',
                 auth_method='ssl', headers=None):

        # Configuration assignments
        self.url = self.config.get('VISA', 'url')
        self.version = self.config.get('VISA', 'version')
        self.user_id = self.config.get('VISA', 'username')
        self.password = self.config.get('VISA', 'password')
        self.cert = os.path.join(self.BASE_DIR, self.config.get('VISA', 'cert'))
        self.key = os.path.join(self.BASE_DIR, self.config.get('VISA', 'key'))
        self.enable_exceptions = self.config.getboolean('VISA', 'enable_exceptions')
        # Optional parameters
        try:
            self.shared_secret = self.config.get('VISA', 'shared_secret')
        except parser.NoOptionError:
            self.shared_secret = ''

        # API path structure: https://domain/resource/api/version/method
        # eg https://sandbox.api.visa.com/cybersource/payments/v1/authorizations
        self.resource = resource
        self.api = api
        self.method = method
        self.api_endpoint = self.url + "/" + "/" + self.resource + "/" + self.api + "/" + self.version + "/" + self.method
        self.http_verb = http_verb

        if query_string:
            self.query_string = query_string
        else:
            self.query_string = ''

        if data:
            self.data = self._obj_to_json(data)
        else:
            self.data = ''

        # Headers setup
        # FIXME: Hardcoded request/response in json format
        self.headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'X-Client-Transaction-ID': self._get_x_client_transaction_id()
        }
        if headers:
            self.headers.update(headers)

        # Session initialization
        self.session = requests.Session()

        # Authentication setup
        if auth_method == 'ssl':
            # certificate-based authentication
            self.session.cert = (self.cert, self.key)
        elif auth_method == 'token':
            # x-pay-token authentication
            x_pay_token = self._get_x_pay_token()
            token_header = {'X-PAY-TOKEN': x_pay_token}
            self.headers.update(token_header)
        else:
            raise AttributeError("auth_method argument value must be either 'ssl' or 'token' ('%s' passed)" % auth_method)

    def send(self):
        """Submits a data object or query string id to VISA using `self.api_endpoint` field and corresponding
        http verb.

        :return: result: Resulting dictionary.
        """

        url = self.api_endpoint + self.query_string

        # Using requests.Session and requests.PreparedRequest to construct request to VDP
        req = requests.Request(
            method=self.http_verb,
            url=url,
            data=self.data,
            auth=(self.user_id, self.password),
            headers=self.headers
        )
        prepped_req = req.prepare()
        result = self.session.send(prepped_req)

        return self._response(result)

    def _response(self, result):
        """Processes a response from Visa Direct API.

        Depending on HTTP code in response, either returns a result or raises corresponding app-level exception.

        :param requests.Response result: Required. Response from VISA
        :return: dict result: Resulting dictionary
        :raises: VisaTimeoutError, VisaDuplicateTransactionError, VisaMessageValidationError, VisaUnauthenticatedError,
                VisaAccessDeniedError, VisaNotFoundError, VisaGeneralError
        """

        code = result.status_code

        if self.enable_exceptions:
            if code == 200:
                result = {
                    'code': code,
                    'endpoint': self.api_endpoint,
                    'http_verb': self.http_verb,
                    'message': result.json()
                }
            else:
                raise self.ERROR_CODES.get(code, VisaGeneralError)(result=result)
        else:
            if code == 202:
                message = result.content
            else:
                message = result.json()

            result = {
                'request': {
                    'endpoint': self.api_endpoint,
                    'http_verb': self.http_verb,
                    'data': self.data,
                },
                'response': {
                    'code': code,
                    'message': message,
                }
            }

        return result



    @staticmethod
    def _get_x_client_transaction_id():
        """Generates random 12-digits value used for X-Client-Transaction-ID header.

        :return: Value for X-Client-Transaction-ID header
        """
        return ''.join(random.choice(string.digits) for _ in range(12))

    def _get_x_pay_token(self):
        """Generates payload for X-PAY-TOKEN header (token-based authentication)
        https://developer.visa.com/guides/vdpguide#two_way_ssl

        :return: X-PAY-TOKEN value
        """
        ts = str(int(time.time()))
        key = self.shared_secret
        resource_path = self.api + '/' + self.version + '/' + self.method
        message = ts + resource_path + self.query_string + self.data

        digest = hmac.new(str.encode(key), msg=str.encode(message), digestmod=hashlib.sha256).hexdigest()

        return "xv2:" + ts + ":" + digest

    @staticmethod
    def _obj_to_json(data):
        """Serializes Python object to json

        :param object data: Python object
        :return: JSON string
        """
        json = jsonpickle.encode(data, unpicklable=False)
        return json



import os

from datetime import datetime

try:
    import configparser as parser
except ImportError:
    import ConfigParser as parser

# Current directory
BASE_DIR = os.path.dirname(__file__)

config = parser.ConfigParser()
config_file = os.path.join(BASE_DIR, '../configuration.ini')

if os.path.exists(config_file):
    config.read(config_file)
else:
    # use example config as a stub for tests
    config.read(os.path.join(BASE_DIR, '../configuration.ini.example'))

transaction_datetime = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")


class VisaDirectTransaction(object):
    """Abstract VISA Transaction object model.

    Constructs and populates general attributes, such as systemsTraceAuditNumber (stan), retrievalReferenceNumber
    (rrn) and localTransactionDateTime.

    :param str stan: **Required**. Systems Trace Audit Number (STAN), 6 digits integer.
    """
    def __init__(self, stan):

        self.systemsTraceAuditNumber = stan
        self.localTransactionDateTime = transaction_datetime
        self.retrievalReferenceNumber = self.get_rrn()

    def get_rrn(self):
        """Generates RRN (retrievalReferenceNumber).

        :return: str rrn: Last digit of current year + number of current day in the year + current hour + STAN
        """
        now = datetime.now()
        year_last_digit = now.strftime('%Y')[3]
        day_of_the_year = now.strftime('%j')
        current_hour = now.strftime('%H')
        return year_last_digit + day_of_the_year + current_hour + str(self.systemsTraceAuditNumber)


class VisaDirectTransactionBatch(object):
    """Abstract TransactionBatch object model.

    Constructs and populates a 'transaction of transactions', normally submitted with `multi` requests.
    This class is not supposed to be instantiated on its own, it is just providing a `shell` for batch requests

    :param list transactions: **Required**. List of :func:`~visa.visadirect.data.VisaDirectTransaction` objects.
    """
    def __init__(self, transactions):
        self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')
        self.acquirerCountryCode = config.get('ACQUIRER', 'acquirer_country_code')
        self.localTransactionDateTime = transaction_datetime
        self.request = transactions


class CardAcceptor(object):
    """Card Acceptor object model.

    https://en.wikipedia.org/wiki/ISO_8583

    :param str name: **Required**. Card Acceptor name.
    :param str terminal_id: **Required**. Card Acceptor terminal ID.
    :param str id_code: **Required**. Card Acceptor ID code.
    :param dict address: **Conditional**. Card acceptor address dictionary. See :func:`CardAcceptorAddress` attributes
        for details.
    """

    ATTR_MAPPINGS = {
        'name': 'name',
        'terminal_id': 'terminalId',
        'id_code': 'idCode',
    }

    def __init__(self, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        try:
            self.address = self.CardAcceptorAddress(**kwargs['address'])
        except KeyError:
            # If address is missing, skip silently, errors will be handled by VDP
            pass

    class CardAcceptorAddress(object):
        """Card Acceptor address object model (part of :func:`~visa.visadirect.data.CardAcceptor` object).

        Populated from `address` dictionary argument for :func:`~visa.visadirect.data.CardAcceptor` class.

        :param str country: **Required**. Address country ISO code.
        :param str zip_code: **Conditional**. Address ZIP code.
        :param str state: **Condtional**. Address state name.
        """
        ATTR_MAPPINGS = {
            'country': 'country',
            'zip_code': 'zipCode',
            'state': 'state',
            'city': 'city'
        }

        def __init__(self, **kwargs):
            self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)


class OriginalDataElements(object):
    """Object model for Original Data Elements object, used for reverse transactions.

    :param int stan: **Required**. Systems trace audit number, 6 digits integer.
    :param datetime transmission_datetime: **Required**. Original AFT Datetime YYYY-MM-DDThh:mm:ss
    :param str approval_code: **Optional**. Approval code from original AFT, 6 characters string.
    """
    ATTR_MAPPINGS = {
        'stan': 'systemsTraceAuditNumber',
        'approval_code': 'approvalCode',
        'transmission_datetime': 'transmissionDateTime',
    }

    def __init__(self, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)
        self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')



import string
import random


class WatchlistObject(object):
    """VisaDirect WatchlistScreening WatchlistInquiry data object model.

    :param str name: **Required**. Name of the person, to receive WLM score. 1-255 characters latin string.
    :param str city: **Required**. Client city, part of :func:`~visa.visadirect.watchlist.data.WatchlistObjectAddress`.
        1-25 characters string.
    :param str issuer_country_code: **Required**. Card issuer 3 characters `ISO country code <https://developer.visa.com/request_response_codes#isoCodes>`_,
        part of :func:`~visa.visadirect.watchlist.data.WatchlistObjectAddress`.

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "840",
                "acquiringBin": "408999",
                "address": {
                    "cardIssuerCountryCode": "USA",
                    "city": "San Francisco"
                },
                "name": "Mohammed Qasim",
                "referenceNumber": "330000550000"
            }
    """
    ATTR_MAPPINGS = {
        'name': 'name'
    }

    def __init__(self, city, issuer_country_code, **kwargs):

        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)
        self.referenceNumber = self.__get_reference()
        self.address = WatchlistObjectAddress(city=city, issuer_country_code=issuer_country_code)

    @staticmethod
    def __get_reference():
        """Generates random 12-digit reference to tie together a WLM score request and WLM score response service calls.

        :return: reference
        """
        return ''.join(random.choice(string.digits) for _ in range(12))


class WatchlistObjectAddress(object):
    """Watchlist address object model, part of :func:`~visa.visadirect.watchlist.data.WatchlistObject`.

    :param str city: **Required**. Client city name.
    :param str issuer_country_code: **Required**. Card issuer 3 characters ISO country code.
    """
    ATTR_MAPPINGS = {
        'city': 'city',
        'issuer_country_code': 'cardIssuerCountryCode'
    }

    def __init__(self, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)



from datetime import datetime


class FundsTransferInquiryData(object):
    """Funds Transfer Inquiry data object model.

    :param str pan: **Required**. Primary account number (PAN). 13-19 characters string.
    :param int stan: **Required**. Systems Trace Audit Number. 6 digits integer. Generated by payment application,
        unique within the scope of application until expired (usually within 24 hours). Can be reused if more than
        999 999 transactions were performed during a day.
    :param str rrn: **Optional**. Retrieval Reference Number. 12 digits string. If not provided, will be generated
        as a function from STAN.
    :param int acquiring_bin: **Optional**. BIN under which FundsTransfer app is registered. 6-11 digits integer.
    :param int acquirer_country_code: **Optional**. 3 digits acquirer country code.

    **Example:**
        ..  code-block:: json

            {
                "primaryAccountNumber": "4957030420210512",
                "retrievalReferenceNumber": "330000550000",
                "systemsTraceAuditNumber": "451006"
            }
    """
    ATTR_MAPPINGS = {
        'pan': 'primaryAccountNumber',
        'rrn': 'retrievalReferenceNumber',
        'acquiring_bin': 'acquiringBin',
        'acquirer_country_code': 'acquirerCountryCode',
    }

    def __init__(self, stan, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        self.systemsTraceAuditNumber = stan

        try:
            self.retrievalReferenceNumber = kwargs['rrn']
        except KeyError:
            self.retrievalReferenceNumber = self._get_rrn()

    def _get_rrn(self):
        """Generates RRN (retrievalReferenceNumber).

        :return: str rrn: Last digit of current year + number of current day in the year + current hour + STAN
        """
        now = datetime.now()
        year_last_digit = now.strftime('%Y')[3]
        day_of_the_year = now.strftime('%j')
        current_hour = now.strftime('%H')
        return year_last_digit + day_of_the_year + current_hour + str(self.systemsTraceAuditNumber)



from pyvdp.visadirect.request import VisaDirectRequest


API = 'mvisa'
METHOD = 'cashoutpushpayments'


def send(transaction):
    """Sends VisaDirect mVISA CashoutPushPayments POST request.

    :param transaction: **Required**. Instance of CashoutPushPaymentTransaction.
    :return: Dictionary with VDP API response.
    """

    c = VisaDirectRequest(api=API, method=METHOD, http_verb='post', data=transaction)
    return c.send()


def get(status_id: str):
    """Sends VisaDirect mVISA CashoutPushPayments GET request.

    :param str status_id: **Required**. Transaction status identifier
    :return: Dictionary with VDP API response
    """

    query_string = '/' + status_id

    c = VisaDirectRequest(api=API, method=METHOD, http_verb='get', query_string=query_string)
    return c.send()



"""
Definitions of data object models for VisaDirect mVISA operations.
"""
from pyvdp.visadirect.data import config, VisaDirectTransaction


class CashinPushPaymentTransaction(VisaDirectTransaction):
    """Data model for mVISA CashinPushPayments API.

    :param int stan: **Required**. Systems trace audit number. 6 digits positive integer.
    :param str recipient_pan: **Required**. Recipient primary account number (PAN). 13-19 characters string.
    :param float amount: **Required**. Transaction amount. 12 digits, 3 fractions.
    :param CardAcceptor card_acceptor: **Required**. Instance of :func:`~visa.visadirect.data.CardAcceptor` data object
    :param str transaction_currency_code: **Required**. Transaction `ISO currency code <https://developer.visa.com/request_response_codes#isoCodes>`_. 3 characters string.
    :param str sender_account_number: **Optional**.
    :param str sender_name: **Optional**. Sender name in Latin characters. Max 30 characters string.

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "643",
                "acquiringBin": "400171",
                "amount": "124.05",
                "businessApplicationId": "CI",
                "cardAcceptor": {
                    "address": {
                        "city": "Bangalore",
                        "country": "IND"
                    },
                    "idCode": "ID-Code123",
                    "name": "Card Accpector ABC"
                },
                "localTransactionDateTime": "2017-03-17T05:42:01",
                "merchantCategoryCode": "4829",
                "recipientPrimaryAccountNumber": "4123640062698797",
                "retrievalReferenceNumber": "430000367618",
                "senderAccountNumber": "4541237895236",
                "senderName": "Mohammed Qasim",
                "senderReference": "1234",
                "systemsTraceAuditNumber": "313042",
                "transactionCurrencyCode": "USD",
                "transactionIdentifier": "381228649430015"
            }

    ..  seealso::

        `VisaDirect mVISA CashinPushPayments Reference <https://developer.visa.com/products/visa_direct/reference#visa_direct__mvisa__v1__cash_in_push_payments>`_

    """
    # Mappings between constructor arguments and VISA API fields
    # {'constructor argument': 'VISA API field name'}
    ATTR_MAPPINGS = {
        'amount': 'amount',
        'recipient_pan': 'recipientPrimaryAccountNumber',
        'transaction_currency_code': 'transactionCurrencyCode',
        'sender_name': 'senderName',
        'sender_account_number': 'senderAccountNumber',
        'card_acceptor': 'cardAcceptor'
    }

    def __init__(self, **kwargs):
        super(CashinPushPaymentTransaction, self).__init__(kwargs['stan'])

        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')
        self.acquirerCountryCode = config.get('ACQUIRER', 'acquirer_country_code')
        self.businessApplicationId = config.get('MVISA', 'cashin_business_application_id')


class CashoutPushPaymentTransaction(VisaDirectTransaction):
    """Data model for mVISA CashoutPushPayments API.

    :param int stan: **Required**. Systems trace audit number. 6 digits positive integer.
    :param str recipient_pan: **Required**. Recipient primary account number (PAN). 13-19 characters string.
    :param float amount: **Required**. Transaction amount. 12 digits, 3 fractions.
    :param CardAcceptor card_acceptor: **Required**. Instance of :func:`~visa.visadirect.data.CardAcceptor` data object
    :param str transaction_currency_code: **Required**. Transaction `ISO currency code <https://developer.visa.com/request_response_codes#isoCodes>`_. 3 characters string.
    :param str sender_account_number: **Optional**.
    :param str sender_name: **Optional**. Sender name in Latin characters. Max 30 characters string.

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "643",
                "acquiringBin": "400171",
                "amount": "124.05",
                "businessApplicationId": "CO",
                "cardAcceptor": {
                    "address": {
                        "city": "mVisa cashout",
                        "country": "IND"
                    },
                    "idCode": "CA-IDCode-77765",
                    "name": "mVisa cashout"
                },
                "localTransactionDateTime": "2017-03-17T05:46:12",
                "merchantCategoryCode": "6012",
                "recipientPrimaryAccountNumber": "4123640062698797",
                "retrievalReferenceNumber": "412123412878",
                "senderAccountNumber": "456789123456",
                "senderName": "Mohammed Qasim",
                "senderReference": "REFNUM123",
                "systemsTraceAuditNumber": "567889",
                "transactionCurrencyCode": "USD",
                "transactionIdentifier": "381228649430015"
            }

    ..  seealso::

        `VisaDirect mVISA CashoutPushPayments Reference <https://developer.visa.com/products/visa_direct/reference#visa_direct__mvisa__v1__cash_out_push_payments_post>`_

    """
    # Mappings between constructor arguments and VISA API fields
    # {'constructor argument': 'VISA API field name'}
    ATTR_MAPPINGS = {
        'amount': 'amount',
        'recipient_pan': 'recipientPrimaryAccountNumber',
        'transaction_currency_code': 'transactionCurrencyCode',
        'sender_name': 'senderName',
        'sender_account_number': 'senderAccountNumber',
        'card_acceptor': 'cardAcceptor'
    }

    def __init__(self, **kwargs):
        super(CashoutPushPaymentTransaction, self).__init__(kwargs['stan'])

        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')
        self.acquirerCountryCode = config.get('ACQUIRER', 'acquirer_country_code')
        self.businessApplicationId = config.get('MVISA', 'cashout_business_application_id')


class MerchantPushPaymentTransaction(VisaDirectTransaction):
    """Data model for mVISA MerchantPushPayments API.

    :param int stan: **Required**. Systems trace audit number. 6 digits positive integer.
    :param str recipient_pan: **Required**. Recipient primary account number (PAN). 13-19 characters string.
    :param float amount: **Required**. Transaction amount. 12 digits, 3 fractions.
    :param CardAcceptor card_acceptor: **Required**. Instance of :func:`~visa.visadirect.data.card_acceptor.CardAcceptor` data object
    :param str transaction_currency_code: **Required**. Transaction `ISO currency code <https://developer.visa.com/request_response_codes#isoCodes>`_. 3 characters string.
    :param PurchaseIdentifier purchase_identifier: **Required**. Instance of :func:`~visa.visadirect.mvisa.data.purchase_identifier.PurchaseIdentifier` data object
    :param str sender_account_number: **Required**. Consumer account number. Max 34 characters string.
    :param str sender_name: **Required**. Sender name in Latin characters. Max 30 characters string.

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "356",
                "acquiringBin": "408972",
                "amount": "124.05",
                "businessApplicationId": "MP",
                "cardAcceptor": {
                    "address": {
                        "city": "KOLKATA",
                        "country": "IND"
                    },
                    "idCode": "CA-IDCode-77765",
                    "name": "Visa Inc. USA-Foster City"
                },
                "feeProgramIndicator": "123",
                "localTransactionDateTime": "2017-03-17T05:51:16",
                "purchaseIdentifier": {
                    "referenceNumber": "REF_123456789123456789123",
                    "type": "1"
                },
                "recipientName": "Jasper",
                "recipientPrimaryAccountNumber": "4123640062698797",
                "retrievalReferenceNumber": "412770451035",
                "secondaryId": "123TEST",
                "senderAccountNumber": "4027290077881587",
                "senderName": "Jasper",
                "senderReference": "",
                "systemsTraceAuditNumber": "451035",
                "transactionCurrencyCode": "INR",
                "transactionIdentifier": "381228649430015"
            }

    ..  seealso::

        `VisaDirect mVISA MerchantPushPayments Reference <https://developer.visa.com/products/visa_direct/reference#visa_direct__mvisa__v1__merchant_push_payments_post>`_

    """

    # Mappings between constructor arguments and VISA API fields
    # {'constructor argument': 'VISA API field name'}
    ATTR_MAPPINGS = {
        'amount': 'amount',
        'recipient_pan': 'recipientPrimaryAccountNumber',
        'transaction_currency_code': 'transactionCurrencyCode',
        'sender_name': 'senderName',
        'sender_account_number': 'senderAccountNumber',
        'purchase_identifier': 'purchaseIdentifier',
        'card_acceptor': 'cardAcceptor',
    }

    def __init__(self, **kwargs):
        super(MerchantPushPaymentTransaction, self).__init__(kwargs['stan'])

        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')
        self.acquirerCountryCode = config.get('ACQUIRER', 'acquirer_country_code')
        self.businessApplicationId = config.get('MVISA', 'mp_business_application_id')


class PurchaseIdentifier(object):
    """Data model for VisaDirect mVISA PurchaseIdentifier object.

    Used as a part of :func:`~visa.visadirect.mvisa.data.MerchantPushPaymentTransaction`.

    :param str type: **Required**. Primary ID as defined by mVISA. 1 character string.
    :param str reference_number: **Required**. Key data element for matching a message to others within a given
        transaction set. Value will be the same as what has been provided in the request. Max 13 characters string.
    """

    ATTR_MAPPINGS = {
        'type': 'type',
        'reference_number': 'referenceNumber',
    }

    def __init__(self, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)



class PavTransaction(object):
    """Visa Payment Account Validation Object data object model.

    :param int stan: **Required**. Systems Trace Audit Number (STAN), 6-digits integer.
    :param str pan: **Required**. Primary Account Number (PAN), 13-19 digits string.
    :param str expiry_date: **Required**. Card expiration date, YYYY-MM.
    :param str cvv2: **Required**. Card CVV2 value. 3-4 characters string.

    **Example:**
        ..  code-block:: json

            {
                "addressVerificationResults": {
                    "postalCode": "T4B 3G5",
                    "street": "2881 Main Street Sw"
                },
                "cardAcceptor": {
                    "address": {
                        "city": "fostr city",
                        "country": "PAKISTAN",
                        "county": "CA",
                        "state": "CA",
                        "zipCode": "94404"
                    },
                    "idCode": "111111",
                    "name": "rohan",
                    "terminalId": "123"
                },
                "cardCvv2Value": "672",
                "cardExpiryDate": "2018-06",
                "primaryAccountNumber": "4957030000313108",
                "retrievalReferenceNumber": "015221743720",
                "systemsTraceAuditNumber": "743720"
            }
    """
    ATTR_MAPPINGS = {
        'stan': 'systemsTraceAuditNumber',
        'pan': 'primaryAccountNumber',
        'expiry_date': 'cardExpiryDate',
        'cvv2': 'cardCvv2Value'
    }

    def __init__(self, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)



class VisaGeneralError(Exception):
    """General exception, raised by VDP API.

    This class is inherited throughout module and implements general exception handling, such as error messages
    display.

    :param requests.Response result: Required. Response from VDP.
    """
    def __init__(self, result):

        if result.content:
            self.message = result.content.decode('utf-8')
        else:
            self.message = "Unknown VISA Error"

    def __str__(self):
        return self.message


class VisaTimeoutError(VisaGeneralError):
    """Raised when POST request is timed out.

    This exception is somewhat special, since while being a timeout error, it is in fact a standard response in case
    multiple transactions were submitted, e.g. through `pullfunds.send(multi=True)` call.

    It returns 202 HTTP code which indicates, that request is received and being processed under the hood, the content
    is a transaction ID which can be used in `get` requests to get a status on this specific submission.

    .. note::
        * VISA API HTTP Code: 202
        * VISA Error Code: None
    """


class VisaDuplicateTransactionError(VisaGeneralError):
    """Raised when duplicate transaction detected.

    .. note::
        * VISA API HTTP Code: 303
        * VISA Error Code: None
    """


class VisaMessageValidationError(VisaGeneralError):
    """Raised when message validation error occured.

    .. note::
        * VISA API HTTP Code: 400
        * VISA Error Code: 3001
    """


class VisaUnauthenticatedError(VisaGeneralError):
    """Raised when client is not authenticated.

    .. note::
        * VISA API HTTP Code: 401
        * VISA Error Code: 9123, 9124
    """


class VisaInvalidCertificateError(VisaGeneralError):
    """Raised when client presented invalid certificate.

    .. note::
        * VISA API HTTP Code: 401
        * VISA Error Code: 9125
    """


class VisaAccessDeniedError(VisaGeneralError):
    """Raised when access to resource is restricted.

    .. note::
        * VISA API HTTP Code: 403
        * VISA Error Code: 9611
    """


class VisaNotFoundError(VisaGeneralError):
    """Raised when resource not found.

    .. note::
        * VISA API HTTP Code: 404
        * VISA Error Code: None
    """


class VisaInvalidPanError(VisaGeneralError):
    """Raised with invalid PAN.

    .. note::
        * VISA API HTTP Code: 404
        * VISA Error Code: 3001
    """



import time

from datetime import datetime


class MerchantSearchData(object):
    """Visa Merchant Search data object model.

    :param dict header: **Required**. A dictionary with header parameters. See :func:`~visa.merchantsearch.data.MerchantSearchData.MerchantSearchHeader`
        for possible keys.
    :param dict search_attrs: **Required**. A dictionary with search atttributes. See :func:`~visa.merchantsearch.data.MerchantSearchData.MerchantSearchAttrList`
        for possible keys.
    :param list response_attrs: **Required**. A list with attributes (Group Names) to include in response.
    :param dict options: **Required**. A dictionary with search options. See :func:`~visa.merchantsearch.data.MerchantSearchData.MerchantSearchOptions`
        for possible keys.

    **Example:**
        ..  code-block:: json

            {
                "header": {
                    "messageDateTime": "2017-03-18T03:39:08.903",
                    "requestMessageId": "Request_001",
                    "startIndex": "0"
                },
                "searchAttrList": {
                    "merchantName": "cmu edctn materials cntr",
                    "merchantStreetAddress": "802 industrial dr",
                    "merchantCity": "Mount Pleasant",
                    "merchantState": "MI",
                    "merchantPostalCode": "48858",
                    "merchantCountryCode": "840",
                    "merchantPhoneNumber": "19897747123",
                    "merchantUrl": "http://www.emc.cmich.edu",
                    "businessRegistrationId": "386004447",
                    "acquirerCardAcceptorId": "424295031886",
                    "acquiringBin": "476197"
                },
                "responseAttrList": [
                    "GNSTANDARD"
                ],
                "searchOptions": {
                    "maxRecords": "5",
                    "matchIndicators": "true",
                    "matchScore": "true",
                    "proximity": [
                        "merchantName"
                    ],
                    "wildCard": [
                        "merchantName"
                    ]
                }
            }
    """

    def __init__(self, header, search_attrs, response_attrs, options):
        self.header = self.MerchantSearchHeader(**header)
        self.searchAttrList = self.MerchantSearchAttrList(**search_attrs)
        self.responseAttrList = response_attrs
        self.searchOptions = self.MerchantSearchOptions(**options)


    class MerchantSearchHeader(object):
        """MerchantSearch header data object model.

        Part of MerchantSearchData object.

        :param str message_id: **Optional**. Unique string ID for service request. String 50 characters max. Default
            'Request_' + epoch
        :param int startIndex: **Optional**. Starting records index in response.
        """

        def __init__(self, **kwargs):
            self.messageDateTime = self._get_datetime()

            if kwargs and 'message_id' in kwargs:
                self.requestMessageId = kwargs['message_id']
            else:
                self.requestMessageId = self._get_message_id()

            if kwargs and 'start_index' in kwargs:
                self.startIndex = kwargs['start_index']
            else:
                self.startIndex = 0

        @staticmethod
        def _get_datetime():
            date_format = '%Y-%m-%dT%H:%M:%S.%f'
            now = datetime.now()
            return now.strftime(date_format)[:-3]

        @staticmethod
        def _get_message_id():
            return 'Request_' + str(int(time.time()))

    class MerchantSearchAttrList(object):
        """MerchantSearch searchAttrList data object model.

        Part of MerchantSearchData object.

        :param str merchant_name: **Conditional**. Name of the merchant. Optional when any one of VisaMerchantId or
            VisaStoreId or BusinessRegistrationId or MerchantUrl or AcquirerCardAcceptorId is provided.
        :param str merchant_street_address: **Conditional**. Merchant street adress.
        :param str merchant_city: **Conditional**. City of the registered merchant.
        :param str merchant_state: **Conditional**. Merchant state code. 2 characters string.
        :param str merchant_postal_code: **Conditional**. Merchant postal code.
        :param int merchant_country_code: **Conditional**. Merchant country code, mandatory with merchant_name.
        :param int merchant_phone_number: **Conditional**. Merchant phone number.
        :param str merchant_url: **Conditional**. Merchant registered URL. Optional when any one of MerchantName or
            VisaMerchantId or VisaStoreId or BusinessRegistrationId or AcquirerCardAcceptorId is provided.
        :param int visa_merchant_id: **Conditional**. Merchant ID provided by VISA. Optional when any one of MerchantName
            or VisaStoreId or BusinessRegistrationId or MerchantUrl or AcquirerCardAcceptorId is provided. 8 characters
            double.
        :param int visa_store_id: **Conditional**. Merchant store/branch ID, provided by VISA. Optional when any one of
            MerchantName or VisaMerchantId or BusinessRegistrationId or MerchantUrl or AcquirerCardAcceptorId is provided.
            9 characters double.
        :param str business_registration_id: **Conditional**. Merchant business/tax registration ID. Optional when any
            one of MerchantName or VisaMerchantId or VisaStoreId or MerchantUrl or AcquirerCardAcceptorId is provided.
        :param str merchant_url: **Conditional**. Merchant registered URL. Optional when any one of MerchantName or
            VisaMerchantId or VisaStoreId or BusinessRegistrationId or AcquirerCardAcceptorId is provided.
        :param str acquirer_card_acceptor_id: **Conditional**. Acquirer card acceptor ID. Optional when any one of
            MerchantName or VisaMerchantId or VisaStoreId or BusinessRegistrationId or MerchantUrl is provided. 15 digits
            string. Prepend 0 if less than 15 digits.
        :param int acquiring_bin: **Conditional**. Acquirer business identification number. Required when
            AcquirerCardAcceptorId is provided.
        """

        ATTR_MAPPINGS = {
            'merchant_name': 'merchantName',
            'merchant_street_address': 'merchantStreetAddress',
            'merchant_city': 'merchantCity',
            'merchant_state': 'merchantState',
            'merchant_postal_code': 'merchantPostalCode',
            'merchant_country_code': 'merchantCountryCode',
            'merchant_phone_number': 'merchantPhoneNumber',
            'merchant_url': 'merchantUrl',
            'visa_merchant_id': 'visaMerchantId',
            'visa_store_id': 'visaStoreId',
            'business_registration_id': 'businessRegistrationId',
            'acquirer_card_acceptor_id': 'acquirerCardAcceptorId',
            'acquiring_bin': 'acquiringBin',
        }

        def __init__(self, **kwargs):
            self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

    class MerchantSearchOptions(object):
        """MerchantSearch searchOptions data object model.

        Part of MerchantSearch data object.

        :param int max_records: **Optional**. Maximum number of records in response. Default 25.
        :param bool match_indicators: **Optional**. Show request attributes, that match a record.
        :param bool match_score: **Optional**. Add matchScore and order response per matchScore.
        :param list proximity: **Optional**. Proximity search on merchant name. If wildcards are used, proximity is ignored.
        :param list wildcards: **Optional**. Wildcard search on merchant name.
        """

        ATTR_MAPPINGS = {
            'max_records': 'maxRecords',
            'match_indicators': 'matchIndicators',
            'match_score': 'matchScore',
        }

        def __init__(self, proximity=None, wildcards=None, **kwargs):
            self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

            if proximity:
                self.proximity = proximity

            if wildcards:
                self.wildCards = wildcards



from pyvdp.visadirect.request import VisaDirectRequest


API = 'fundstransfer'
METHOD = 'pullfundstransactions'


def send(transaction, multi=False):
    """Submits a PullFundsTransactions request.

    :param transaction: **Required**. Instance of :func:`~visa.visadirect.fundstransfer.data.PullFundsTransaction` or :func:`~visa.visadirect.fundstransfer.data.MultiPullFundsTransaction`.
    :param bool multi: **Conditional**. Indicates that transaction is a batch (:func:`~visa.visadirect.fundstransfer.data.MultiPullFundsTransaction`)
    :return: Dictionary with VDP API response

    **Example:**
        ..  code-block:: python

            from visa.visadirect.data.transaction import PullTransaction
            from visa.visadirect.fundstransfer import pull_funds

            ca = CardAcceptor(name='Acceptor 1',
                              country='RU',
                              terminal_id='TID-9999',
                              id_code='CA-IDCode-77765')

            t = PullFundsTransaction(stan=123456,
                                     amount=123.45,
                                     sender_pan='12345678123456678',
                                     sender_card_expiry_date='12-2020',
                                     sender_currency_code='USD',
                                     card_acceptor=ca)

            result = pull_funds.send(transaction=t)
    """
    method = METHOD

    if multi:
        method = 'multipullfundstransactions'

    c = VisaDirectRequest(api=API, method=method, http_verb='post', data=transaction)
    return c.send()


def get(status_id, multi=False):
    """Fetches a status of previously submitted :func:`~visa.visadirect.fundstransfer.data.PullFundsTransaction` by
    transaction identifier.

    :param str status_id: **Required**. Transaction status identifier.
    :param bool multi: **Conditional**. Indicates that transaction is a batch (:func:`~visa.visadirect.fundstransfer.data.MultiPullFundsTransaction`)
    :return: Dictionary with VDP API response

    **Example:**
        ..  code-block:: python

                from visa.visadirect.data.transaction import PullTransaction, MultiPullTransaction
                from visa.visadirect.fundstransfer import pull_funds
                from visa.exceptions import VisaTimeoutError

                ca = CardAcceptor(name='Acceptor 1',
                                  country='RU',
                                  terminal_id='TID-9999',
                                  id_code='CA-IDCode-77765')

                atom = PullFundsTransaction(stan=123456,
                                            amount=123.45,
                                            sender_pan='12345678123456678',
                                            sender_card_expiry_date='12-2020',
                                            sender_currency_code='USD',
                                            card_acceptor=ca)

                # Let's pretend its a batch
                transactions = [atom]
                mpt = MultiPullFundsTransaction(transactions=transactions)

                try:
                    # This will return HTTP 202 (multi always does)
                    pull_funds.send(transaction=mpt, multi=true)
                except VisaTimeoutError as e:
                    # Grab an exception message and use it as an argument
                    result = pull_funds.get(status_id=e.message)
    """
    method = METHOD

    if multi:
        method = 'multipullfundstransactions'

    query_string = '/' + status_id

    c = VisaDirectRequest(api=API, method=method, http_verb='get', query_string=query_string)
    return c.send()



from pyvdp.visadirect.request import VisaDirectRequest


API = 'reports'
METHOD = 'transactiondata'


def get(from_date, to_date):
    """Fetches transaction report for given dates range.

    :param str from_date: **Required**. Starting date *ddmmyyyy*
    :param str to_date: **Required**. Ending date *ddmmyyyy*
    :return: CSV-formatted list of transactions
    """
    query = '?fromDate=' + from_date + '&toDate=' + to_date
    c = VisaDirectRequest(api=API, method=METHOD, http_verb='get', query_string=query)
    return c.send()



from pyvdp.request import VisaRequest


class VisaMerchantSearchRequest(VisaRequest):
    """Implements HTTP requests to Visa MerchantSearch APIs.

    https://developer.visa.com/products/merchant_search/guides

    :param str query_string: **Conditional**. Query string to append to URL
    :param object data: **Conditional**. Payload for POST/PUT requests
    :param dict headers: **Optional**. Additional headers as dictionary
    """

    def __init__(self, query_string='', data=None, headers=None):

        super(VisaMerchantSearchRequest, self).__init__(resource='merchantsearch',
                                                        api='',
                                                        method='search',
                                                        http_verb='post',
                                                        query_string=query_string,
                                                        data=data,
                                                        headers=headers)




from pyvdp.visadirect.request import VisaDirectRequest


API = 'fundstransfer'
METHOD = 'pushfundstransactions'


def send(transaction, multi=False):
    """Submits a PushFundsTransactions request.

    :param transaction: **Required**. Instance of :func:`~visa.visadirect.fundstransfer.data.PushFundsTransaction` or :func:`~visa.visadirect.fundstransfer.data.MultiPushFundsTransaction`.
    :param bool multi: **Conditional**. Indicates that transaction is a batch (:func:`~visa.visadirect.fundstransfer.data.MultiPushFundsTransaction`).
    :return: Dictionary with VDP API response.

    **Example:**
        ..  code-block:: python

                from visa.visadirect.data.card_acceptor import CardAcceptor
                from visa.visadirect.fundstransfer.data.push_transaction import PushTransaction

                ca = CardAcceptor(name='Acceptor 1',
                                  country='USA',
                                  terminal_id='TID-9999',
                                  id_code='CA-IDCode-77765',
                                  zip_code='12345',
                                  state='CA')

                t = PushFundsTransaction(stan=123456,
                                         amount=123.45,
                                         transaction_currency_code='USD',
                                         card_acceptor=ca,
                                         sender_pan='1234567812345678',
                                         recipient_pan='8765432187654321',
                                         recipient_name='Test')

                result = push_funds.send(transaction=t)
    """
    method = METHOD

    if multi:
        method = 'multipushfundstransactions'

    c = VisaDirectRequest(api=API, method=method, http_verb='post', data=transaction)
    return c.send()


def get(status_id, multi=False):
    """Fetches a status of previously submitted :func:`~visa.visadirect.fundstransfer.data.PushFundsTransaction` by
    transaction identifier.

    :param str status_id: **Required**. Transaction status identifier
    :param bool multi: **Conditional**. Indicates that transaction is a batch (:func:`~visa.visadirect.fundstransfer.data.MultiPushFundsTransaction`)
    :return: Dictionary with VDP API response

    **Example:**
        ..  code-block:: python

                from visa.visadirect.data.card_acceptor import CardAcceptor
                from visa.visadirect.data.transaction import PushTransaction, MultiPushTransaction
                from visa.visadirect.fundstransfer import push_funds
                from visa.exceptions import VisaTimeoutError

                ca = CardAcceptor(name='Acceptor 1',
                                  country='USA',
                                  terminal_id='TID-9999',
                                  id_code='CA-IDCode-77765',
                                  zip_code='12345',
                                  state='CA')

                atom = PushFundsTransaction(stan=123456,
                                            amount=123.45,
                                            transaction_currency_code='USD',
                                            card_acceptor=ca,
                                            sender_pan='1234567812345678',
                                            recipient_pan='8765432187654321',
                                            recipient_name='Test')

                # Let's pretend its a batch
                mpt = [atom]
                t = MultiPushTransaction(transactions=mpt)

                try:
                    # This will return HTTP 202 (multi always does)
                    push_funds.send(transaction=mpt, multi=true)
                except VisaTimeoutError as e:
                    # Grab an exception message and use it as an argument
                    result = push_funds.get(status_id=e.message)
    """
    method = METHOD

    if multi:
        method = 'multipushfundstransactions'

    query_string = '/' + status_id

    c = VisaDirectRequest(api=API, method=method, http_verb='get', query_string=query_string)
    return c.send()



from pyvdp.visadirect.request import VisaDirectRequest


def send(data):
    """Submits a WatchlistInquiry request.

    :param WatchlistObject data: **Required**. Instance of :func:`~visa.visadirect.watchlist.data.WatchlistObject`.
    :return: Dictionary with VDP API response

    **Example:**
        ..  code-block:: python

            from visa.visadirect.watchlist import watchlistinquiry
            from visa.visadirect.watchlist.data import WatchlistObject

            data = WatchlistObject(name='Pavel Pokrovskiy',
                                   city='Moscow',
                                   issuer_country_code='RU')

            result = watchlistinquiry.send(data)
    """
    c = VisaDirectRequest(api='watchlistscreening', method='watchlistinquiry', http_verb='post', data=data)
    return c.send()



from pyvdp.merchantsearch.request import VisaMerchantSearchRequest
from pyvdp.merchantsearch.data import MerchantSearchData


def send(data):
    """Submits a merchant search request.

    :param MerchantSearchData data: **Required**. Instance of :func:`~visa.merchantsearch.data.MerchantSearchData`.
    :return: A response from VDP.

    **Example:**
        ..  code-block:: python

            from visa.merchantsearch.data import MerchantSearchData
            from visa.merchantsearch import search

            # See docs for MerchantSearchAttrList to find out how to build search_attrs
            search_attrs = {
                'merchant_name': 'Acme Inc',
                'visa_merchant_id': 12345678,
                'visa_store_id': 123456789,
                'business_registration_id': 386004447,
                'merchant_url': 'http://www.emc.cmich.edu',
                'acquirer_card_acceptor_id': 424295031886,
                'acquiring_bin': 476197
            }

            response_attrs = ['GNSTANDARD']

            search_options = {
                'max_records': 5,
                'match_indicators': True,
                'match_score': True,
                'proximity': list('Acme'),
                'wildcard': list('*cme*')
            }

            data = MerchantSearchData(search_attrs=search_attrs,
                                      response_attrs=response_attrs,
                                      options=search_options)

            result = search.send(data=data)
    """
    c = VisaMerchantSearchRequest(data=data)
    return c.send()



from pyvdp.visadirect.request import VisaDirectRequest


API = 'fundstransfer'
METHOD = 'reversefundstransactions'


def send(transaction, multi=False):
    """Submits a ReverseFundsTransactions request.

    :param ReverseTransaction|MultiReverseTransaction transaction: **Required**. Instance of :func:`~visa.visadirect.fundstransfer.data.PushFundsTransaction` or
        :func:`~visa.visadirect.fundstransfer.data.MultiPushFundsTransaction`.
    :param bool multi: **Conditional**. Indicates that transaction is a batch (:func:`~visa.visadirect.fundstransfer.data.MultiReverseFundsTransaction`).
    :return: Dictionary with VDP response

    **Example:**
        ..  code-block:: python

            from visa.visadirect.data.card_acceptor import CardAcceptor
            from visa.visadirect.data.original_data_elements import OriginalDataElements
            from visa.visadirect.fundstransfer.data.reverse_transaction import ReverseTransaction

            from visa.visadirect.fundstransfer import reverse_funds

            ca = CardAcceptor(name='Acceptor 1',
                              country='RU',
                              terminal_id='TID-9999',
                              id_code='CA-IDCode-77765')

            ode = OriginalDataElements(stan=123456,
                                       approval_code='20304B',
                                       transmission_datetime='2017-02-16T12:59:26')

            t = ReverseTransaction(stan=123456,
                                   ode=ode,
                                   sender_pan='1234567812345678',
                                   sender_currency_code='USD',
                                   amount=123.45,
                                   sender_card_expiry_date='02-2020',
                                   transaction_identifier='123456,
                                   card_acceptor=ca)

            result = reverse_funds.send(transaction=t)
    """
    method = METHOD

    if multi:
        method = 'multireversefundstransactions'

    c = VisaDirectRequest(api=API, method=method, http_verb='post', data=transaction)
    return c.send()


def get(status_id, multi=False):
    """Fetches a status of previously submitted :func:`~visa.visadirect.fundstransfer.data.ReverseFundsTransaction` by
    transaction identifier.

    :param str status_id: **Required**. Transaction status identifier.
    :param bool multi: **Conditional**. Indicates that transaction is a batch (:func:`~visa.visadirect.fundstransfer.data.MultiReverseFundsTransaction`)
    :return: Dictionary with VDP API response

    **Example:**
        ..  code-block:: python

            from visa.visadirect.data.card_acceptor import CardAcceptor
            from visa.visadirect.data.original_data_elements import OriginalDataElements

            from visa.visadirect.fundstransfer.data.reverse_transaction import ReverseFundsTransaction, MultiReverseFundsTransaction

            from visa.visadirect.fundstransfer import reverse_funds

            ca = CardAcceptor(name='Acceptor 1',
                              country='RU',
                              terminal_id='TID-9999',
                              id_code='CA-IDCode-77765')

            ode = OriginalDataElements(stan=123456,
                                       approval_code='20304B',
                                       transmission_datetime='2017-02-16T12:59:26')

            atom = ReverseTransaction(stan=123456,
                                      ode=ode,
                                      sender_pan='1234567812345678',
                                      sender_currency_code='USD',
                                      amount=123.45,
                                      sender_card_expiry_date='02-2020',
                                      transaction_identifier=12345,
                                      card_acceptor=ca)

            mrt = [atom]
            t = MultiReverseTransaction(transactions=mrt)

            try:
                # This will return HTTP 202 (multi always does)
                reverse_funds.send(transaction=t, multi=True)
            except VisaTimeoutError as e:
                # Grab an exception message and use it as an argument
                result = reverse_funds.get(status_id=e.message)
    """
    method = METHOD

    if multi:
        method = 'multireversefundstransactions'

    query_string = '/' + status_id

    c = VisaDirectRequest(api=API, method=method, http_verb='get', query_string=query_string)
    return c.send()



from pyvdp.request import VisaRequest


class PaaiRequest(VisaRequest):
    """Implements HTTP requests to Payment Account Attributes Inquiry APIs.

    https://developer.visa.com/products/paai/guides

    :param str http_verb: **Required**. VDP API HTTP verb (GET, POST).
    :param str query_string: **Conditional**. Query string to append to URL.
    :param object data: **Conditional**. Payload for POST/PUT requests.
    :param dict headers: **Optional**. Additional headers as dictionary.
    """

    def __init__(self, api, method, http_verb, query_string='', data=None, headers=None):
        super(PaaiRequest, self).__init__(resource='paai',
                                          api=api,
                                          method=method,
                                          http_verb=http_verb,
                                          query_string=query_string,
                                          data=data,
                                          headers=headers)



from pyvdp.visadirect.data import config, VisaDirectTransaction, VisaDirectTransactionBatch


class PullFundsTransaction(VisaDirectTransaction):
    """VisaDirect FundsTransfer PullFunds data object model.

    :param int stan: **Required**. systemsTraceAuditNumber, 6 digits integer.
    :param float amount: **Required**. Transaction amount nnn.mmm.
    :param str sender_pan: **Required**. Sender PAN (Primary Account Number), 13-19 characters.
    :param datetime sender_card_expiry_date: **Required**. Sender card expiration date, YYYY-MM.
    :param str sender_currency_code: **Required**. Sender currency code, 3 characters ISO currency code.
    :param bool multi: **Conditional**. If True, this transaction is a part of batch. See :func:`~visa.visadirect.fundstransfer.data.MultiPullFundsTransaction`

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "840",
                "acquiringBin": "408999",
                "amount": "124.02",
                "businessApplicationId": "AA",
                "cardAcceptor": {
                    "address": {
                        "country": "USA",
                        "county": "San Mateo",
                        "state": "CA",
                        "zipCode": "94404"
                    },
                    "idCode": "ABCD1234ABCD123",
                    "name": "Visa Inc. USA-Foster City",
                    "terminalId": "ABCD1234"
                    },
                "cavv": "0700100038238906000013405823891061668252",
                "foreignExchangeFeeTransaction": "11.99",
                "localTransactionDateTime": "2017-03-17T08:20:42",
                "retrievalReferenceNumber": "330000550000",
                "senderCardExpiryDate": "2015-10",
                "senderCurrencyCode": "USD",
                "senderPrimaryAccountNumber": "4895142232120006",
                "surcharge": "11.99",
                "systemsTraceAuditNumber": "451001"
            }

    """
    ATTR_MAPPINGS = {
        'amount': 'amount',
        'sender_pan': 'senderPrimaryAccountNumber',
        'sender_card_expiry_date': 'senderCardExpiryDate',
        'sender_currency_code': 'senderCurrencyCode',
        'card_acceptor': 'cardAcceptor'
    }

    def __init__(self, multi=False, **kwargs):
        super(PullFundsTransaction, self).__init__(kwargs['stan'])

        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        # Acquirer details are only populated for individual pull transactions, not batches
        if not multi:
            self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')
            self.acquirerCountryCode = config.get('ACQUIRER', 'acquirer_country_code')
            self.businessApplicationId = config.get('FUNDSTRANSFER', 'business_application_id')


class MultiPullFundsTransaction(VisaDirectTransactionBatch):
    """VisaDirect FundsTransfer MultiPullFunds data object model.

    It is container object, which incorporate PullFundsTransaction objects under the same acquirer
    credentials.

    :param list transactions: **Required**. List of :func:`~visa.visadirect.fundstransfer.data.PullFundsTransaction` objects

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "608",
                "acquiringBin": "408999",
                "businessApplicationId": "AA",
                "localTransactionDateTime": "2017-03-17T09:11:08",
                "merchantCategoryCode": "6012",
                "request": [
                    {
                        "amount": "100.00",
                        "cardAcceptor": {
                            "address": {
                                "country": "USA",
                                "county": "00",
                                "state": "CA",
                                "zipCode": "94454"
                            },
                            "idCode": "5678",
                            "name": "Mr Smith",
                            "terminalId": "1234"
                        },
                        "cavv": "0700020718799100000002980179911000000000",
                        "localTransactionDateTime": "2017-03-17T09:11:08",
                        "retrievalReferenceNumber": "401010101011",
                        "senderCardExpiryDate": "2020-12",
                        "senderCurrencyCode": "USD",
                        "senderPrimaryAccountNumber": "4895140000066666",
                        "systemsTraceAuditNumber": "101011"
                    },
                    {
                        "amount": "100.00",
                        "cardAcceptor": {
                            "address": {
                                "country": "USA",
                                "county": "00",
                                "state": "CA",
                                "zipCode": "94454"
                            },
                            "idCode": "5678",
                            "name": "Mr Smith",
                            "terminalId": "1234"
                        },
                        "cavv": "0700020718799100000002980179911000000000",
                        "localTransactionDateTime": "2017-03-17T09:11:08",
                        "retrievalReferenceNumber": "401010101011",
                        "senderCardExpiryDate": "2020-12",
                        "senderCurrencyCode": "USD",
                        "senderPrimaryAccountNumber": "4895140000066666",
                        "systemsTraceAuditNumber": "101011"
                    }
                ]
            }
    """
    def __init__(self, transactions):
        super(MultiPullFundsTransaction, self).__init__(transactions=transactions)
        self.businessApplicationId = config.get('FUNDSTRANSFER', 'business_application_id')


class PushFundsTransaction(VisaDirectTransaction):
    """VisaDirect FundsTransfer PushFunds data object model.

    :param int stan: **Required**. Systems Trace Audit Number (STAN), 6 digits integer.
    :param float amount: **Required**. Transaction amount, nnn.mmm.
    :param str transaction_currency_code: **Required**. Transaction currency ISO code, 3 characters.
    :param str sender_pan: **Required**. Sender primary account number (PAN), 13-19 digits.
    :param str recipient_pan: **Required**. Recipient primary account number (PAN), 13-19 digits.
    :param str recipient_name: **Required**. Recipient name.
    :param str sender_name: **Conditional**. Sender name.
    :param str sender_address: **Conditional**. Sender address.
    :param str sender_city: **Conditional**. Sender city.
    :param str sender_country_code: **Conditional**. Sender ISO country code, 3 characters.
    :param str sender_state_code: **Conditional**. Sender country state code.
    :param bool multi: **Conditional**. If True, constructs object, used as a part of batch. See :func:`~visa.visadirect.fundstransfer.data.MultiPushFundsTransaction`.

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "840",
                "acquiringBin": "408999",
                "amount": "124.05",
                "businessApplicationId": "AA",
                "cardAcceptor": {
                    "address": {
                        "country": "USA",
                        "county": "San Mateo",
                        "state": "CA",
                        "zipCode": "94404"
                    },
                    "idCode": "CA-IDCode-77765",
                    "name": "Visa Inc. USA-Foster City",
                    "terminalId": "TID-9999"
                },
                "localTransactionDateTime": "2017-03-17T09:01:14",
                "merchantCategoryCode": "6012",
                "pointOfServiceData": {
                    "motoECIIndicator": "0",
                    "panEntryMode": "90",
                    "posConditionCode": "00"
                },
                "recipientName": "rohan",
                "recipientPrimaryAccountNumber": "4957030420210462",
                "retrievalReferenceNumber": "412770451018",
                "senderAccountNumber": "4957030420210454",
                "senderAddress": "901 Metro Center Blvd",
                "senderCity": "Foster City",
                "senderCountryCode": "124",
                "senderName": "Mohammed Qasim",
                "senderReference": "",
                "senderStateCode": "CA",
                "sourceOfFundsCode": "05",
                "systemsTraceAuditNumber": "451018",
                "transactionCurrencyCode": "USD",
                "transactionIdentifier": "381228649430015"
            }
    """

    ATTR_MAPPINGS = {
        'amount': 'amount',
        'transaction_currency_code': 'transactionCurrencyCode',
        'sender_pan': 'senderAccountNumber',
        'sender_name': 'senderName',
        'sender_address': 'senderAddress',
        'sender_city': 'senderCity',
        'sender_country_code': 'senderCountryCode',
        'sender_state_code': 'senderStateCode',
        'recipient_pan': 'recipientPrimaryAccountNumber',
        'recipient_name': 'recipientName',
        'card_acceptor': 'cardAcceptor'
    }

    def __init__(self, multi=False, **kwargs):
        super(PushFundsTransaction, self).__init__(kwargs['stan'])

        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)

        # Acquirer details are only populated for individual push transactions, not batches
        if not multi:
            self.acquiringBin = config.get('ACQUIRER', 'acquiring_bin')
            self.acquirerCountryCode = config.get('ACQUIRER', 'acquirer_country_code')
            self.businessApplicationId = config.get('FUNDSTRANSFER', 'business_application_id')


class MultiPushFundsTransaction(VisaDirectTransactionBatch):
    """VisaDirect FundsTransfer MultiPushFunds data object model.

    It is container object, which incorporate PushFundsTransaction objects under the same acquirer
    credentials.

    :param list transactions: **Required**. List of :func:`~visa.visadirect.fundstransfer.data.PushFundsTransaction` objects

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "840",
                "acquiringBin": "408999",
                "businessApplicationId": "AA",
                "localTransactionDateTime": "2017-03-17T09:12:46",
                "merchantCategoryCode": "6012",
                "request": [
                    {
                        "amount": "100.00",
                        "cardAcceptor": {
                            "address": {
                                "country": "USA",
                                "county": "00",
                                "state": "CA",
                                "zipCode": "94454"
                            },
                            "idCode": "5678",
                            "name": "Mr Smith",
                            "terminalId": "1234"
                        },
                        "feeProgramIndicator": "123",
                        "localTransactionDateTime": "2017-03-17T09:12:46",
                        "recipientName": "Akhila",
                        "recipientPrimaryAccountNumber": "4957030420210454",
                        "retrievalReferenceNumber": "401010101011",
                        "senderAccountNumber": "4005520000011126",
                        "senderAddress": "My Address",
                        "senderCity": "My City",
                        "senderCountryCode": "USA",
                        "senderName": "Mr Name",
                        "senderReference": "",
                        "senderStateCode": "CA",
                        "sourceOfFundsCode": "01",
                        "systemsTraceAuditNumber": "101011",
                        "transactionCurrencyCode": "USD",
                        "transactionIdentifier": "234234234234234"
                    },
                    {
                        "amount": "100.00",
                        "cardAcceptor": {
                            "address": {
                                "country": "USA",
                                "county": "00",
                                "state": "CA",
                                "zipCode": "94454"
                            },
                            "idCode": "5678",
                            "name": "Mr Smith",
                            "terminalId": "1234"
                        },
                        "feeProgramIndicator": "123",
                        "localTransactionDateTime": "2017-03-17T09:12:46",
                        "recipientName": "Akhila",
                        "recipientPrimaryAccountNumber": "4957030420210454",
                        "retrievalReferenceNumber": "401010101012",
                        "senderAccountNumber": "4840920103511221",
                        "senderAddress": "My Address",
                        "senderCity": "My City",
                        "senderCountryCode": "USA",
                        "senderName": "Mr Name",
                        "senderReference": "",
                        "senderStateCode": "CA",
                        "sourceOfFundsCode": "01",
                        "systemsTraceAuditNumber": "101012",
                        "transactionCurrencyCode": "USD",
                        "transactionIdentifier": "234234234234234"
                    }
                ]
            }
    """
    def __init__(self, transactions):
        super(MultiPushFundsTransaction, self).__init__(transactions=transactions)
        self.businessApplicationId = config.get('FUNDSTRANSFER', 'business_application_id')


class ReverseFundsTransaction(PullFundsTransaction):
    """VisaDirect FundsTransfer ReverseFunds data object model.

    :param int stan: **Required**. Systems Trace Audit Number, required, 6 digits integer.
    :param str sender_pan: **Required**. Sender Primary Account Number, 13-19 characters sender PAN.
    :param str sender_card_expiry_date: **Required**. Sender Card Expiry Date,, YYYY-MM sender card expiration date.
    :param str sender_currency_code: **Required**. Sender Currency Code, required, 3 characters ISO currency code.
    :param float amount: **Required**. Transaction amount.
    :param int transaction_identifier: **Required**. Transaction Identifier, 15 digits integer.
    :param OriginalDataElements ode: **Required**. Instance of :func:`~visa.visadirect.data.OriginalDataElements` object.

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "608",
                "acquiringBin": "408999",
                "amount": "24.01",
                "cardAcceptor": {
                    "address": {
                        "country": "USA",
                        "county": "San Mateo",
                        "state": "CA",
                        "zipCode": "94404"
                    },
                    "idCode": "VMT200911026070",
                    "name": "Visa Inc. USA-Foster City",
                    "terminalId": "365539"
                },
                "localTransactionDateTime": "2017-03-17T09:02:32",
                "originalDataElements": {
                    "acquiringBin": "408999",
                    "approvalCode": "20304B",
                    "systemsTraceAuditNumber": "897825",
                    "transmissionDateTime": "2017-03-17T09:02:32"
                },
                "pointOfServiceCapability": {
                    "posTerminalEntryCapability": "2",
                    "posTerminalType": "4"
                },
                "pointOfServiceData": {
                    "motoECIIndicator": "0",
                    "panEntryMode": "90",
                    "posConditionCode": "00"
                },
                "retrievalReferenceNumber": "330000550000",
                "senderCardExpiryDate": "2015-10",
                "senderCurrencyCode": "USD",
                "senderPrimaryAccountNumber": "4895100000055127",
                "systemsTraceAuditNumber": "451050",
                "transactionIdentifier": "381228649430011"
            }
    """
    ATTR_MAPPINGS = {
        'sender_pan': 'senderPrimaryAccountNumber',
        'sender_card_expiry_date': 'senderCardExpiryDate',
        'sender_currency_code': 'senderCurrencyCode',
        'amount': 'amount',
        'transaction_identifier': 'transactionIdentifier',
        'card_acceptor': 'cardAcceptor',
        'ode': 'originalDataElements'
    }

    def __init__(self, ode, **kwargs):
        super(ReverseFundsTransaction, self).__init__(**kwargs)
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)
        self.originalDataElements = ode


class MultiReverseFundsTransaction(MultiPullFundsTransaction):
    """VisaDirect FundsTransfer MultiReverseFunds data object model.

    It is container object, which incorporate ReverseFundsTransaction objects under the same acquirer
    credentials.

    :param list transactions: **Required**. List of :func:`~visa.visadirect.fundstransfer.data.ReverseFundsTransaction` objects

    **Example:**
        ..  code-block:: json

            {
                "acquirerCountryCode": "840",
                "acquiringBin": "408999",
                "localTransactionDateTime": "2017-03-17T09:09:05",
                "request": [
                    {
                        "amount": "100.00",
                        "cardAcceptor": {
                            "address": {
                                "country": "USA",
                                "county": "00",
                                "state": "CA",
                                "zipCode": "94454"
                            },
                            "idCode": "5678",
                            "name": "Mr Smith",
                            "terminalId": "1234"
                        },
                        "localTransactionDateTime": "2017-03-17T09:09:05",
                        "originalDataElements": {
                            "acquiringBin": "408999",
                            "approvalCode": "1ABCDE",
                            "systemsTraceAuditNumber": "228112",
                            "transmissionDateTime": "2017-03-17T09:09:05"
                        },
                        "retrievalReferenceNumber": "401010101011",
                        "senderCardExpiryDate": "2020-12",
                        "senderCurrencyCode": "USD",
                        "senderPrimaryAccountNumber": "4485810000000131",
                        "systemsTraceAuditNumber": "101011",
                        "transactionIdentifier": "101010101010"
                    },
                    {
                        "amount": "100.00",
                        "cardAcceptor": {
                            "address": {
                                "country": "USA",
                                "county": "00",
                                "state": "CA",
                                "zipCode": "94454"
                            },
                            "idCode": "5678",
                            "name": "Mr Smith",
                            "terminalId": "1234"
                        },
                        "localTransactionDateTime": "2017-03-17T09:09:05",
                        "originalDataElements": {
                            "acquiringBin": "408999",
                            "approvalCode": "1ABCDE",
                            "systemsTraceAuditNumber": "228112",
                            "transmissionDateTime": "2017-03-17T09:09:05"
                        },
                        "retrievalReferenceNumber": "401010101011",
                        "senderCardExpiryDate": "2020-12",
                        "senderCurrencyCode": "USD",
                        "senderPrimaryAccountNumber": "4485810000000131",
                        "systemsTraceAuditNumber": "101011",
                        "transactionIdentifier": "101010101010"
                    }
                ]
            }
    """
    def __init__(self, transactions):
        super(MultiReverseFundsTransaction, self).__init__(transactions=transactions)
        self.businessApplicationId = config.get('FUNDSTRANSFER', 'business_application_id')




from setuptools import setup, find_packages

DESCRIPTION = 'A collection of wrappers for Visa Developer Program APIs (https://developer.visa.com/)'


CLASSIFIERS = [
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
]

setup(
    name='pyvdp',
    version='1.1.6',
    author='Pavel Pokrovskiy',
    author_email='ppokrovsky@gmail.com',
    license='MIT',
    description=DESCRIPTION,
    keywords=['visa', 'vdp', 'payment'],
    platforms=['OS Independent'],
    classifiers=CLASSIFIERS,
    test_suite='tests',
    packages=find_packages('.', exclude=['tests']),
    install_requires=[
        'setuptools',
        'six',
        'requests',
        'jsonpickle'
    ],
    url='https://www.github.com/ppokrovsky/pyvdp',
    include_package_data=True,
    package_data={'': ['LICENSE.txt', 'README.md'], 'pyvdp': ['configuration.ini.example']},
)



from pyvdp.pav.request import VisaPavRequest
from pyvdp.pav.data import PavTransaction


def send(data: PavTransaction):
    """Submits a payment account validation request.

    :param PavTransaction data: **Required**. Instance of :func:`~visa.pav.data.PavTransaction`.
    :return: A response from VDP.

    **Example:**
        ..  code-block:: python

            from visa.pav.data import PavTransaction
            from visa.pav import cardvalidation

            data = PavTransaction(stan=123456,
                                  pan='1234567812345678',
                                  expiry_date='2020-04',
                                  cvv2='123')

            result = cardvalidation.send(data=data)
    """
    c = VisaPavRequest(method='cardvalidation', http_verb='post', data=data)
    return c.send()



class GeneralInquiryData(object):
    """General Card Attributes Inquiry data object model.

    :param str pan: **Required**. Primary account number (PAN). 13-19 characters string.

    **Example:**
        ..  code-block:: json

            {
                "primaryAccountNumber": "4465390000029077"
            }
    """
    ATTR_MAPPINGS = {
        'pan': 'primaryAccountNumber'
    }

    def __init__(self, **kwargs):
        self.__dict__.update((self.ATTR_MAPPINGS[k], v) for k, v in kwargs.items() if k in self.ATTR_MAPPINGS and v)



from pyvdp.paai.request import PaaiRequest


API = 'fundstransferattinq'
METHOD = 'cardattributes/fundstransferinquiry'


def send(data):
    """Submits Funds Transfer Inquiry request.

    :param data: **Required**. Instance of :func:`~visa.paai.fundstransferattinq.cardattributes.data.FundsTransferInquiryData`
    :return: Dictionary with VDP response

    **Example:**
        ..  code-block:: python

            from visa.paai.fundstransferattinq.cardattributes.data import FundsTransferInquiryData
            from visa.paai.fundstransferattinq.cardattribute import fundstransferinquiry

            data = FundsTransferInquiryData(stan=123456,
                                            rrn="330000123456",
                                            pan="4957030420210512")

            result = fundstransferinquiry.send(data=data)
    """
    c = PaaiRequest(api=API, method=METHOD, http_verb='POST', data=data)
    return c.send()



from pyvdp.request import VisaRequest


class VisaDirectRequest(VisaRequest):
    """Implements HTTP requests to Visa Direct APIs.

    https://developer.visa.com/products/visa_direct/guides

    :param str api: **Required**. VDP API name
    :param str method: **Required**. VDP API method name
    :param str http_verb: **Required**. VDP API HTTP verb (GET, POST)
    :param str query_string: **Conditional**. Query string to append to URL
    :param object data: **Conditional**. Payload for POST/PUT requests
    :param dict headers: **Optional**. Additional headers as dictionary
    """

    def __init__(self, api, method, http_verb, query_string='', data=None, headers=None):

        super(VisaDirectRequest, self).__init__(resource='visadirect',
                                                api=api,
                                                method=method,
                                                http_verb=http_verb,
                                                query_string=query_string,
                                                data=data,
                                                headers=headers)




from pyvdp.paai.request import PaaiRequest


API = 'generalattinq'
METHOD = 'cardattributes/generalinquiry'


def send(data):
    """Submits General Card Attributes Inquiry request.

    :param data: **Required**. Instance of :func:`~visa.paai.generalattinq.cardattributes.data.GeneralInquiryData`
    :return: Dictionary with VDP response

    **Example:**
        ..  code-block:: python

            from visa.paai.generalattinq.cardattributes.data import GeneralInquiryData
            from visa.paai.generalattinq.cardattributes import generalinquiry

            data = GeneralInquiryData(pan="4957030420210512")

            result = generalinquiry.send(data=data)
    """
    c = PaaiRequest(api=API, method=METHOD, http_verb='POST', data=data)
    return c.send()



from pyvdp.request import VisaRequest


class VisaPavRequest(VisaRequest):
    """Instantiates connection to PaymentAccountValidation VDP API endpoint.

    :param str method: **Required**. VDP method name.
    :param str http_verb: **Required**. HTTP verb (GET, POST).
    :param str query_string: **Conditional**. Query string to append to URL.
    :param object data: **Conditional**. Object to submit as POST/PUT payload.
    :param dict headers: **Optional**. Additional headers as dictionary.
    """
    def __init__(self, method, http_verb, query_string='', data=None, headers=None):
        super(VisaPavRequest, self).__init__(resource='pav', api='', method=method, http_verb=http_verb,
                                             query_string=query_string, data=data, headers=headers)
